/* ccgo: go/gtk/gtk.cc
 * 
 * Copyright 2004 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <gtk.hh>
#include <go/sgf.hh>
#include <go/debug.hh>

#include <gtkmm/box.h>
#include <gtkmm/frame.h>
#include <gtkmm/stock.h>
#include <gtkmm/menubar.h>
#include <gtkmm/alignment.h>
#include <gtkmm/separator.h>
#include <gtkmm/spinbutton.h>
#include <gtkmm/scrolledwindow.h>
#include <sstream>
#include <fstream>

using namespace gtk;

// members of class MarkWin

MarkWin::MarkWin() :
	Gtk::Window(Gtk::WINDOW_TOPLEVEL)
{
	Gtk::Frame * f = Gtk::manage(new Gtk::Frame);
	f->set_shadow_type(Gtk::SHADOW_OUT);
	add(* f);
	Gtk::VBox * vb = Gtk::manage(new Gtk::VBox);
	f->add(* vb);
	vb->pack_start(* Gtk::manage(new Gtk::Label("Use Mark:")));
	Gtk::RadioButton::Group gp;
	rbs.push_back(Gtk::manage(new Gtk::RadioButton(gp, "None")));
	rbs.back()->signal_clicked().connect(bind(slot(* this, & MarkWin::select_mark), ccgo::MARK_NONE));
	vb->pack_start(* rbs.back());
	rbs.push_back(Gtk::manage(new Gtk::RadioButton(gp, "Circle")));
	rbs.back()->signal_clicked().connect(bind(slot(* this, & MarkWin::select_mark), ccgo::MARK_CIRCLE));
	vb->pack_start(* rbs.back());
	rbs.push_back(Gtk::manage(new Gtk::RadioButton(gp, "Triangle")));
	rbs.back()->signal_clicked().connect(bind(slot(* this, & MarkWin::select_mark), ccgo::MARK_TRIANGLE));
	vb->pack_start(* rbs.back());
	rbs.push_back(Gtk::manage(new Gtk::RadioButton(gp, "Square")));
	rbs.back()->signal_clicked().connect(bind(slot(* this, & MarkWin::select_mark), ccgo::MARK_SQUARE));
	vb->pack_start(* rbs.back());
	rbs.push_back(Gtk::manage(new Gtk::RadioButton(gp, "Cross Mark")));
	rbs.back()->signal_clicked().connect(bind(slot(* this, & MarkWin::select_mark), ccgo::MARK_CROSS));
	vb->pack_start(* rbs.back());
	rbs.push_back(Gtk::manage(new Gtk::RadioButton(gp, "Select")));
	rbs.back()->signal_clicked().connect(bind(slot(* this, & MarkWin::select_mark), ccgo::MARK_SELECT));
	vb->pack_start(* rbs.back());
	rbs.push_back(Gtk::manage(new Gtk::RadioButton(gp, "White Territory")));
	rbs.back()->signal_clicked().connect(bind(slot(* this, & MarkWin::select_mark), ccgo::MARK_W_TERRITORY));
	vb->pack_start(* rbs.back());
	rbs.push_back(Gtk::manage(new Gtk::RadioButton(gp, "Black Territory")));
	rbs.back()->signal_clicked().connect(bind(slot(* this, & MarkWin::select_mark), ccgo::MARK_B_TERRITORY));
	vb->pack_start(* rbs.back());

	vb->pack_start(* Gtk::manage(new Gtk::HSeparator));

	// entry
	vb->pack_start(* Gtk::manage(new Gtk::Label("Use Label:")));
	vb->pack_start(et);
	et.set_width_chars(3);
	et.signal_activate().connect(slot(* this, & MarkWin::enter_label));

	vb->pack_start(* Gtk::manage(new Gtk::HSeparator));
	Gtk::Button * bt = Gtk::manage(new Gtk::Button(Gtk::Stock::CLOSE));
	bt->signal_clicked().connect(slot(* this, & Gtk::Window::hide));
	vb->pack_end(* bt);
}

void MarkWin::show_win(int x, int y, std::string loc)
{
	if (loc == "") set_title("Marking");
	else set_title(std::string("Marking at ") + loc);
	show_all();
	int xx;
	int yy;
	get_size(xx, yy);
	move(x - xx / 2, y - yy / 2);
}

void MarkWin::set_mark(ccgo::Mark marking)
{
	m = marking;
	rbs[m]->set_active();
	l = "";
	et.set_text(l);
}

void MarkWin::set_label(std::string label)
{
	m = ccgo::MARK_NONE;
	rbs[m]->set_active();
	l = label;
	et.set_text(l);
}

ccgo::Mark MarkWin::get_mark()
{
	return m;
}

const std::string & MarkWin::get_label()
{
	return l;
}

void MarkWin::select_mark(ccgo::Mark marking)
{
	if (! rbs[marking]->get_active()) return;
	m = marking;
	l = "";
	hide();
}

void MarkWin::enter_label()
{
	msg(DBG_DEBUG) << "label = " << et.get_text() << '\n';
	l = et.get_text();
	m = ccgo::MARK_NONE;
	hide();
}

// members of class View
View::View() :
	b(0),
	dp("---"),
	sv_w(310),
	sv_h(350),
	sv_p(250),
	mode(MODE_VIEW),
	persist_mode(mode),
	mk_loc(ccgo::MOVE_NONE),
	terri(0),
	my_game(0)
{
	set_title("gtk");
	set_default_size(sv_w, sv_h);
	Gtk::VBox * vb = Gtk::manage(new Gtk::VBox);
	add(* vb);

	// board
	b = new gtk::Board;
	b->set_board(this);
	b->clicked.connect(slot(* this, & View::board_clicked));

	// create menu bar
	Gtk::MenuBar * mb = Gtk::manage(new Gtk::MenuBar);
	vb->pack_start(* mb, Gtk::PACK_SHRINK);

	// File menu
	Gtk::Menu * mm = Gtk::manage(new Gtk::Menu);
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("_Open", slot(* this, & View::file_open)));
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("_Save", slot(* this, & View::file_save)));
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("_Close", Gtk::Menu::AccelKey("<control>w"), slot(* this, & View::hide)));
	mb->items().push_back(Gtk::Menu_Helpers::MenuElem("_File", * mm));

	// Game menu
	mm = Gtk::manage(new Gtk::Menu);
	Gtk::Menu * init_menu = Gtk::manage(new Gtk::Menu);
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("_Initialize", * init_menu));
	m_init = & mm->items().back();
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("Game _Mode", m_mode));

	mm->items().push_back(Gtk::Menu_Helpers::SeparatorElem());
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("_Pass", slot(* this, & View::game_pass)));
	mm->items().push_back(Gtk::Menu_Helpers::SeparatorElem());
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("Create _Non-Move Node", Gtk::Menu::AccelKey("<control>n"), slot(* this, & View::create_node)));
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("_Delete Node Branch", Gtk::Menu::AccelKey("<control>d"), slot(* this, & View::delete_moves)));
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("_Clear Marks", slot(* this, & View::clear_marks)));
	mb->items().push_back(Gtk::Menu_Helpers::MenuElem("_Game", * mm));

	// Game->Init menu
	for (int i = 2; i < 10; i ++) {
		std::string n = "_0 Handicap Stones";
		n[1] += i;
		init_menu->items().push_back(Gtk::Menu_Helpers::MenuElem(n, bind(slot(* this, & View::game_handicap), i)));
	}
	init_menu->items().push_back(Gtk::Menu_Helpers::SeparatorElem());
	init_menu->items().push_back(Gtk::Menu_Helpers::MenuElem("_Clear Handicaps", bind(slot(* this, & View::game_handicap), - 1)));
	m_init->set_sensitive(false);

	// Game->Mode menu
	Gtk::RadioMenuItem::Group mode_group;
	m_mode.items().push_back(Gtk::Menu_Helpers::RadioMenuElem(mode_group, "_View Only", bind(slot(* this, & View::game_mode), MODE_VIEW)));
	m_mode.items().push_back(Gtk::Menu_Helpers::RadioMenuElem(mode_group, "Handicap _Initialization", bind(slot(* this, & View::game_mode), MODE_INIT)));
	m_mode.items().push_back(Gtk::Menu_Helpers::RadioMenuElem(mode_group, "_Play Moves", bind(slot(* this, & View::game_mode), MODE_PLAY)));
	m_mode.items().push_back(Gtk::Menu_Helpers::RadioMenuElem(mode_group, "Place _Marks", bind(slot(* this, & View::game_mode), MODE_MARK)));
	m_mode.items().push_back(Gtk::Menu_Helpers::RadioMenuElem(mode_group, "_Setup Board", bind(slot(* this, & View::game_mode), MODE_SETUP)));
	m_mode.items().push_back(Gtk::Menu_Helpers::RadioMenuElem(mode_group, "Sco_re Board", bind(slot(* this, & View::game_mode), MODE_SCORE)));

	m_mode.items()[MODE_INIT].set_sensitive(false);

	m_mode.set_active(mode);

	// View menu
	mm = Gtk::manage(new Gtk::Menu);
	Gtk::Menu_Helpers::CheckMenuElem * cme;
	// Mark Last Move
	cme = new Gtk::Menu_Helpers::CheckMenuElem("_Mark Last Move", Gtk::Menu::AccelKey("<control>m"), slot(* this, & View::toggle_mark_last));
	ml = Glib::RefPtr<Gtk::CheckMenuItem>::cast_dynamic(cme->get_child());
	mm->items().push_back(* cme);
	delete cme;
	// Show Ko
	cme = new Gtk::Menu_Helpers::CheckMenuElem("Show _Ko", Gtk::Menu::AccelKey("<control>k"), slot(* this, & View::toggle_show_ko));
	sk = Glib::RefPtr<Gtk::CheckMenuItem>::cast_dynamic(cme->get_child());
	mm->items().push_back(* cme);
	mm->items().push_back(Gtk::Menu_Helpers::MenuElem("Board", b->get_view_menu()));
	// Together...
	mb->items().push_back(Gtk::Menu_Helpers::MenuElem("_View", * mm));

	vb->pack_end(vp);

	Gtk::HBox * hb = Gtk::manage(new Gtk::HBox);
	Gtk::Frame * fr = Gtk::manage(new Gtk::Frame);
	fr->set_shadow_type(Gtk::SHADOW_IN);
	vp.add1(* fr);
	fr->add(* hb);

	hb->pack_start(* b);

	// control box
	Gtk::VBox * vb2 = Gtk::manage(new Gtk::VBox);
	hb->pack_end(* vb2, Gtk::PACK_SHRINK);

	Gtk::Alignment * al = Gtk::manage(new Gtk::Alignment(0.5, 0.5, 0, 0));
	// map button
	al->add(btM);
	vb2->pack_start(* al, Gtk::PACK_SHRINK);

	// move up and down
	Gtk::SpinButton * sb = Gtk::manage(new Gtk::SpinButton());
	sb->set_increments(- 1, - 10);
	sb->set_numeric();
	prog = sb->get_adjustment();
	prog->set_lower(0);
	prog->signal_value_changed().connect(slot(* this, & View::depth_change));
	vb2->pack_start(* sb, Gtk::PACK_SHRINK);

	// move number
	vb2->pack_start(dp, Gtk::PACK_SHRINK);

	// left and right
	al = Gtk::manage(new Gtk::Alignment(0.5, 0.5, 0, 0));
	vb2->pack_start(* al, Gtk::PACK_SHRINK);
	hb = Gtk::manage(new Gtk::HBox);
	al->add(* hb);
	hb->pack_start(btL, Gtk::PACK_SHRINK);
	// hb->pack_start(btM, Gtk::PACK_SHRINK);
	hb->pack_end(btR, Gtk::PACK_SHRINK);
	btL.add(* Gtk::manage(new Gtk::Image(Gtk::Stock::GO_BACK, Gtk::ICON_SIZE_SMALL_TOOLBAR)));
	// btM.add(* Gtk::manage(new Gtk::Label("M")));
	btM.add(daM);
	btR.add(* Gtk::manage(new Gtk::Image(Gtk::Stock::GO_FORWARD, Gtk::ICON_SIZE_SMALL_TOOLBAR)));
	btL.signal_clicked().connect(slot(* this, & View::alternative_left));
	btR.signal_clicked().connect(slot(* this, & View::alternative_right));
	btM.signal_toggled().connect(slot(* this, & View::map_toggle));
	daM.signal_expose_event().connect(slot(* this, & View::map_handle_draw));

	// information
	vb2->pack_start(* Gtk::manage(new Gtk::Label("score")), Gtk::PACK_SHRINK);
	al = Gtk::manage(new Gtk::Alignment(0.5, 0.5, 0, 0));
	vb2->pack_start(* al, Gtk::PACK_SHRINK);
	hb = Gtk::manage(new Gtk::HBox);
	al->add(* hb);
	hb->pack_start(* Gtk::manage(new Gtk::Label("B:")), Gtk::PACK_SHRINK);
	hb->pack_start(b_cap, Gtk::PACK_SHRINK);
	al = Gtk::manage(new Gtk::Alignment(0.5, 0.5, 0, 0));
	vb2->pack_start(* al, Gtk::PACK_SHRINK);
	hb = Gtk::manage(new Gtk::HBox);
	al->add(* hb);
	hb->pack_start(* Gtk::manage(new Gtk::Label("W:")), Gtk::PACK_SHRINK);
	hb->pack_start(w_cap, Gtk::PACK_SHRINK);

	// mode options

	vb2->pack_start(o_mode, Gtk::PACK_SHRINK);
	Gtk::Menu * om = Gtk::manage(new Gtk::Menu);
	o_mode.set_menu(* om);
	om->items().push_back(Gtk::Menu_Helpers::MenuElem("View  "));
	om->items().push_back(Gtk::Menu_Helpers::MenuElem("Init"));
	om->items().push_back(Gtk::Menu_Helpers::MenuElem("Play"));
	om->items().push_back(Gtk::Menu_Helpers::MenuElem("Mark"));
	om->items().push_back(Gtk::Menu_Helpers::MenuElem("Set"));
	om->items().push_back(Gtk::Menu_Helpers::MenuElem("Score"));
	o_mode.set_history(mode);
	o_mode.signal_changed().connect(bind(slot(* this, & View::game_mode), MODE_UNKNOWN));

	// turn options
	al = Gtk::manage(new Gtk::Alignment(0.5, 0.5, 0, 0));
	al->add(o_turn);
	vb2->pack_start(* al, Gtk::PACK_SHRINK);
	om = Gtk::manage(new Gtk::Menu);
	o_turn.set_menu(* om);
	om->items().push_back(Gtk::Menu_Helpers::MenuElem("B"));
	om->items().push_back(Gtk::Menu_Helpers::MenuElem("W"));
	o_turn.set_history(get_turn() - 1);
	o_turn.signal_changed().connect(slot(* this, & View::game_turn_change));
	o_turn.set_sensitive(mode == MODE_SETUP);

	// text box
	Gtk::ScrolledWindow * sw = Gtk::manage(new Gtk::ScrolledWindow);
	sw->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_ALWAYS);
	sw->add(tv);
	tv.set_wrap_mode(Gtk::WRAP_WORD);
	fr = Gtk::manage(new Gtk::Frame);
	fr->set_shadow_type(Gtk::SHADOW_IN);
	vp.pack2(* fr, false, true);
	fr->add(* sw);
	fr->signal_size_allocate().connect(slot(* this, & View::paned_moved));

	// game map
	mw.signal_hide().connect(slot(* this, & View::map_hide));
	mw.set_transient_for(* this);
	mw.set_modal(true);
	mw.gm.node_select.connect(slot(* this, & View::select_node));

	// markwin
	mkw.set_transient_for(* this);
	mkw.set_modal(true);
	mkw.signal_hide().connect(slot(* this, & View::mark_set));

	// file dialog
	file_win.set_transient_for(* this);
	file_win.set_modal(true);
	// file_win.signal_response().connect(slot(* this, & View::file_selected));

	vp.set_position(sv_p);
}

View::~View()
{
	if (where()) Walk::set_node(0);
	if (b) delete b;
	if (terri) delete terri;
	if (my_game) {
		delete my_game;
	}
}

void View::set_node(ccgo::Node * node)
{
	msg(DBG_DEBUG) << "setting node\n";
	if (where()) do_old_node();
	Walk::set_node(node);
	b->set_board(this);
	b->update();
	// make path
	path.clear();
	ccgo::Node * n = where();
	if (! n) return;
	while (n->up()) {
		path.insert(path.begin(), n->get_index());
		n = n->up();
	}
	std::string t = "ccgo: ";
	set_title(t + where()->game()->get_name());
	do_new_node();
}

void View::on_realize()
{
	Gtk::Window::on_realize();
	ml->set_active();
	sk->set_active();
	msg(DBG_DEBUG) << "View realized\n";

	// sv_w = get_width();
	// sv_h = get_height();
	// sv_p = vp.get_position();
}

bool View::on_configure_event(GdkEventConfigure * event)
{
	if (! Gtk::Window::on_configure_event(event)) return false;
	int w;
	int h;
	create_pango_layout("999")->get_pixel_size(w, h);
	daM.set_size_request(w, h);
	// btM.resize_children();
	// btM.check_resize();
	return true;
}

void View::on_size_allocate(GtkAllocation * allocation)
{
	Gtk::Window::on_size_allocate(allocation);
	if (allocation->width != sv_w || allocation->height != sv_h) {
		sv_w = allocation->width;
		sv_h = allocation->height;
	}
}

void View::clear()
{
	msg(DBG_DEBUG) << "clearing view\n";
	Walk::clear();
	b->update();
}

void View::set_state(int loc, State state)
{
	Walk::set_state(loc, state);
	b->update(loc);
}

void View::set_ko(const std::set<int> & ko)
{
	Walk::set_ko(ko);
	for (std::set<int>::const_iterator i = ko.begin(); i != ko.end(); i ++) b->update(* i);
}

void View::toggle_mark_last()
{
	if (where()->get_move() == ccgo::MOVE_NONE) return;
	if (ml->get_active()) b->add_mark(where()->get_move(), ccgo::MARK_CIRCLE);
	else b->add_mark(where()->get_move(), ccgo::MARK_NONE);
}

void View::toggle_show_ko()
{
	const std::set<int> & k = get_ko();
	ccgo::Mark m = sk->get_active() ? ccgo::MARK_CROSS :  ccgo::MARK_NONE;
	for (std::set<int>::const_iterator i = k.begin(); i != k.end(); i ++) b->add_mark(* i, m);
}

void View::set_max_depth()
{
	msg(DBG_DEBUG) << "setting max depth\n";
        int r = get_level();
	ccgo::Node * n = where();
	while (n->down_num()) {
		if (r < path.size()) n = n->down(path[r]);
		else {
			n = n->down(0);
			path.push_back(0);
		}
		r ++;
	}
	prog->set_upper(r);
	std::ostringstream s;
	s << r;
	dp.set_text(s.str());
}

void View::depth_change()
{
	int d = int(prog->get_value());
	if (get_level() == d) return;

	msg(DBG_DEBUG) << "changing level to " << d << "\n";
	do_old_node();

	while (get_level() > d) up();
	while (get_level() < d) down(path[get_level()]);

	do_new_node();
}

void View::set_alternative()
{
	// if (! is_realized()) return;
	int i = where()->get_index();
	btL.set_sensitive(i > 0);
	btR.set_sensitive(i >= 0 && i + 1 < where()->up()->down_num());
}

void View::alternative_left()
{
	int i = where()->get_index();
	if (i < 1) return;

	do_old_node();

	up();
	path.resize(get_level());
	down(i - 1);
	path.push_back(i - 1);

	do_new_node();
}

void View::alternative_right()
{
	int i = where()->get_index();
	if (i < 0 || i + 1 >= where()->up()->down_num()) return;

	do_old_node();

	up();
	path.resize(get_level());
	down(i + 1);
	path.push_back(i + 1);

	do_new_node();
}

void View::board_clicked(int loc)
{
	int i;
	msg(DBG_DEBUG) << "board clicked at " << loc << '\n';

	switch (mode) {
	case MODE_VIEW:
		break;
	case MODE_INIT:
		game_add_handi(loc);
		break;
	case MODE_PLAY:
		i = game_add_move(loc);
		if (i < 0) {
			msg(DBG_INFO) << "move not allowed\n";
			break;
		}
		do_old_node();

		path.resize(get_level());
		down(i);
		path.push_back(i);

		do_new_node();
		break;
	case MODE_MARK:
		{
			int x;
			int y;
			get_pointer(x, y);
			int x0;
			int y0;
			get_window()->get_origin(x0, y0);
			mk_loc = loc;


			ccgo::Marking * mk = dynamic_cast<ccgo::Marking *>(where()->get_prop(ccgo::Marking::the_id()));
			if (mk) mkw.set_mark(mk->get_mark(loc));
			else mkw.set_mark(ccgo::MARK_NONE);

			ccgo::Labels * lb = dynamic_cast<ccgo::Labels *>(where()->get_prop(ccgo::Labels::the_id()));
			if (lb) {
				const std::string & s = lb->get_label(loc);
				if (s != "") mkw.set_label(s);
			}
			b->add_mark(loc, ccgo::MARK_SELECT);
			mkw.show_win(x0 + x, y0 + y, loc_to_str(loc));
		}
		break;
	case MODE_SETUP:
		game_set_site(loc, ccgo::Board::State((Board::get_state(loc) + 1) % 3));
		draw_map_handle();
		break;
	case MODE_SCORE:
		{
			assert(Board::get_state(loc));
			std::vector<int> g;
			dead_group(loc, g);
			Terri t;
			ccgo::Mark m;
			if (terri[loc] == TERRI_NEUTRAL) {
				t = Board::get_state(loc) == STATE_BLACK ? TERRI_WHITE : TERRI_BLACK;
				m = t == TERRI_BLACK ? ccgo::MARK_B_TERRITORY : ccgo::MARK_W_TERRITORY;
			} else {
				t = TERRI_NEUTRAL;
				m = ccgo::MARK_NONE;
			}
			for (std::vector<int>::iterator i = g.begin(); i != g.end(); i ++) {
				terri[* i] = t;
			}
			msg(DBG_INFO) << (m ? "mark dead " : "mark alive ") << g.size() << " stones\n";
			do_score();
		}
		break;
	}
}

void View::delete_moves()
{
	if (! where()->up()) return;
	int i = where()->get_index();

	do_old_node();

	up();
	path.resize(get_level());
	where()->del_child(i);

	do_new_node();
}

void View::create_node()
{
	int i;
	i = game_add_setup();
	do_old_node();
	path.resize(get_level());
	down(i);
	path.push_back(i);
	do_new_node();
}

void View::do_old_node()
{
	switch (mode) { // ending some modes
	case MODE_SCORE:
		m_mode.items()[persist_mode].activate();
		break;
	}
	b->clear_marks();
	b->clear_labels();
	b->clear_blocks();
	std::string s = tv.get_buffer()->get_text();
	if (s == "") where()->set_prop(ccgo::Comment::the_id(), 0);
	else {
		ccgo::Comment * c = new ccgo::Comment;
		c->set_comment(s);
		where()->set_prop(ccgo::Comment::the_id(), c);
	}
	// remove empty Marking prop
	ccgo::Marking * mk = dynamic_cast<ccgo::Marking *>(where()->get_prop(ccgo::Marking::the_id()));
	if (mk && ! mk->get_marks().size()) {
		where()->set_prop(ccgo::Marking::the_id(), 0);
	}
	// remove empty Labels prop
	ccgo::Labels * lb = dynamic_cast<ccgo::Labels *>(where()->get_prop(ccgo::Labels::the_id()));
	if (lb && ! lb->get_labels().size()) {
		where()->set_prop(ccgo::Labels::the_id(), 0);
	}
}

void View::do_new_node()
{
	set_max_depth();
	prog->set_value(get_level());
	set_alternative();
	b->set_track(gtk::Board::TrackMode(get_turn()));

	switch (mode) {
	case MODE_VIEW:
		b->set_track(gtk::Board::TRACK_NONE);
		break;
	case MODE_INIT:
		b->set_track(gtk::Board::TRACK_BLACK);
		break;
	case MODE_PLAY:
		b->set_track(gtk::Board::TrackMode(get_turn()));
		break;
	case MODE_MARK:
		b->set_track(gtk::Board::TRACK_ALL);
		break;
	case MODE_SETUP:
		if (where()->get_move() != ccgo::MOVE_NONE || where()->down_num()) { // can not setup
			m_mode.items()[persist_mode].activate();
		} else b->set_track(gtk::Board::TRACK_ALL);
		break;
	case MODE_SCORE:
		// impossible
		break;
	}

	// last move
	if (ml->get_active() && where()->get_move() >= 0) b->add_mark(where()->get_move(), ccgo::MARK_CIRCLE);
	// ko
	if (sk->get_active()) {
		const std::set<int> & k = get_ko();
		for (std::set<int>::const_iterator i = k.begin(); i != k.end(); i ++) {
			b->add_mark(* i, ccgo::MARK_CROSS);
		}
	}
	ccgo::Comment * c = (ccgo::Comment *)(where()->get_prop(ccgo::Comment::the_id()));
	if (c) {
		tv.get_buffer()->set_text(c->get_comment());
	} else {
		tv.get_buffer()->set_text("");
	}

	ccgo::Labels * l = (ccgo::Labels *)(where()->get_prop(ccgo::Labels::the_id()));
	if (l) {
		const std::map<int, std::string> & lb  = l->get_labels();
		for (std::map<int, std::string>::const_iterator i = lb.begin(); i != lb.end(); i ++) {
			msg(DBG_DEBUG) << i->first << ':' << i->second << "\n";
			b->add_label(i->first, i->second);
		}
	}
	double bs = 0;
	double ws = 0;
	ccgo::Marking * m = (ccgo::Marking *)(where()->get_prop(ccgo::Marking::the_id()));
	if (m) {
		const std::map<int, ccgo::Mark> & mk = m->get_marks();
		for (std::map<int, ccgo::Mark>::const_iterator i = mk.begin(); i != mk.end(); i ++) {
			b->add_mark(i->first, i->second);
			// counting territory
			switch (i->second) {
			case ccgo::MARK_B_TERRITORY:
				switch (Board::get_state(i->first)) {
				case STATE_WHITE:
					bs ++;
					break;
				case STATE_BLACK:
					ws ++;
					break;
				}
				bs ++;
				break;
			case ccgo::MARK_W_TERRITORY:
				switch (Board::get_state(i->first)) {
				case STATE_WHITE:
					bs ++;
					break;
				case STATE_BLACK:
					ws ++;
					break;
				}
				ws ++;
				break;
			}
		}
	}
	const std::set<int> & ko = get_ko();
	for (std::set<int>::const_iterator i = ko.begin(); i != ko.end(); i ++) {
		b->set_block(* i, true);
	}

	draw_map_handle();

	if (game_can_init()) {
		m_init->set_sensitive(true);
		m_mode.items()[MODE_INIT].set_sensitive(true);
		o_mode.get_menu()->items()[MODE_INIT].set_sensitive(true);
	} else {
		if (mode == MODE_INIT) m_mode.items()[persist_mode].activate();
		m_init->set_sensitive(false);
		m_mode.items()[MODE_INIT].set_sensitive(false);
		o_mode.get_menu()->items()[MODE_INIT].set_sensitive(false);
	}
	o_turn.set_history(get_turn() - 1);

	// information
	std::ostringstream s;
	bs += get_b_cap();
	ws += get_w_cap();
	double km = where()->game()->get_komi();
	if (km > 0) ws += km;
	else bs -= km;
	s << bs;
	b_cap.set_text(s.str());
	s.str("");
	s << ws;
	w_cap.set_text(s.str());
}

void View::paned_moved(GtkAllocation * allocation)
{
	if (vp.get_position() == sv_p) return;
	sv_p = vp.get_position();
	msg(DBG_DEBUG) << "paned moved: height = " << allocation->height
		       << "; position = " << sv_p << "\n";
}

void View::map_toggle()
{
	if (btM.get_active()) {
		int x;
		int y;
		int x1;
		int y1;
		mw.show_map(where(), get_level(), & x1, & y1);
		// reposition the window
		daM.get_window()->get_origin(x, y);
		GtkAllocation r = daM.get_allocation();
		x += r.width / 2;
		y += r.height / 2;
		mw.move_map(x - x1, y - y1);
	} else mw.hide();
}

void View::map_hide()
{
	if (btM.get_active()) btM.set_active(false);
}

bool View::map_handle_draw(GdkEventExpose * event)
{
	draw_map_handle();
	return true;
}

void View::draw_map_handle()
{
	if (! daM.is_realized()) return;
	GtkAllocation r = daM.get_allocation();
	int cx = r.width / 2;
	int cy = r.height / 2;
	daM.get_window()->clear_area(0, 0, r.width, r.height);
	if (where()->get_move() != ccgo::MOVE_NONE) { // played node
		std::string s;
		if (where()->get_move() >= 0) s = loc_to_str(where()->get_move());
		else s = "pa.";
		Glib::RefPtr<Pango::Layout> t = daM.create_pango_layout(s);
		int w;
		int h;
		t->get_pixel_size(w, h);
		if (get_turn() == ccgo::Position::TURN_BLACK) {
			daM.get_window()->draw_rectangle(daM.get_style()->get_white_gc(), true, cx - w / 2 - 1, cy - h / 2 - 1, w + 1, h + 1);
			daM.get_window()->draw_layout(daM.get_style()->get_black_gc(), cx - w / 2, cy - h / 2, t);
		} else {
			daM.get_window()->draw_rectangle(daM.get_style()->get_black_gc(), true, cx - w / 2 - 1, cy - h / 2 - 1, w + 1, h + 1);
			daM.get_window()->draw_layout(daM.get_style()->get_white_gc(), cx - w / 2, cy - h / 2, t);
		}
	} else if (where()->get_delta()) { // setup
		Glib::RefPtr<Pango::Layout> t = daM.create_pango_layout("-/+");
		int w;
		int h;
		t->get_pixel_size(w, h);
		daM.get_window()->draw_layout(daM.get_style()->get_fg_gc(daM.get_state()), cx - w / 2, cy - h / 2, t);
	} else if (where()->up() == 0) { // start
		std::string s = "[S]";
		ccgo::Game * g = dynamic_cast<ccgo::Game *>(where());
		assert(g);
		if (g->get_init() && g->get_init()->get_turn() == ccgo::Position::TURN_WHITE) {
			s = "[H]";
		}
		Glib::RefPtr<Pango::Layout> t = daM.create_pango_layout(s);
		int w;
		int h;
		t->get_pixel_size(w, h);
		daM.get_window()->draw_layout(daM.get_style()->get_fg_gc(daM.get_state()), cx - w / 2, cy - h / 2, t);		
	}
}

void View::select_node(ccgo::Node * node)
{
	mw.hide();

	assert(node->game() == where()->game());
	if (node == where()) return;
	// move to node
	// find path to node
	std::vector<int> np;
	ccgo::Node * n = node;
	while(n->up()) {
		np.insert(np.begin(), n->get_index());
		n = n->up();
	}
	int i = 0;
	while (i < np.size() && np[i] == path[i]) {
		n = n->down(np[i]);
		i ++;
	}
	do_old_node();

	while (get_level() > i) up();
	if (i == np.size()) {  // new node is on path
		while (get_level() < i) down(path[get_level()]);
	} else {
		i = np.size();
		while (get_level() < i) down(np[get_level()]);
		path = np;
	}

	do_new_node();
}

void View::file_open()
{
	file_win.set_title("Opening SGF file");
	file_win.show_all();
	file_win.complete("*.sgf");
	int result = file_win.run();
	file_win.hide();
	if (result == Gtk::RESPONSE_OK) {
		std::string s = file_win.get_filename();
		msg(DBG_INFO) << "opening SGF file " << s << '\n';
		std::ifstream f(s.c_str()); // , std::ios::in);
		ccgo::Game * g = 0;
		if (f && (g = ccgo::load_sgf(f))) {
			set_node(g);
			if (my_game) delete my_game;
			my_game = g;
		} else {
			msg(DBG_ERROR) << "fail to load SGF file " << s << '\n';
		}
	}
}

void View::file_save()
{
	file_win.set_title("Saving SGF file");
	file_win.show_all();
	file_win.complete("*.sgf");
	int result = file_win.run();
	file_win.hide();
	if (result == Gtk::RESPONSE_OK) {
		std::string s = file_win.get_filename();
		msg(DBG_INFO) << "saving SGF file " << s << '\n';
		std::ofstream f(s.c_str()); // , std::ios::out);
		if (f) {
			ccgo::save_sgf(where()->game(), f);
		} else {
			msg(DBG_ERROR) << "can open file " << s << " for SGF save \n";
		}
	}
}

void View::game_pass()
{
	int i = game_add_move(ccgo::MOVE_PASS);
	if (i < 0) {
		msg(DBG_INFO) << "pass not allowed\n";
		return;
	}
	if (mode != MODE_PLAY) m_mode.items()[persist_mode].activate();
	do_old_node();

	path.resize(get_level());
	down(i);
	path.push_back(i);

	do_new_node();
}

void View::game_handicap(int num_stones)
{
	assert(game_can_init());
	game_set_handi(num_stones);
	do_new_node();
}

void View::game_mode(Mode new_mode)
{
	Mode old_mode = mode;
	if (new_mode == MODE_UNKNOWN) {
		new_mode = (Mode) o_mode.get_history();
		if (new_mode == mode) return;
		Gtk::RadioMenuItem * r = dynamic_cast<Gtk::RadioMenuItem *>(& m_mode.items()[new_mode]);
		assert(r);
		mode = new_mode;
		if (! r->get_active()) r->set_active();
	} else { // from menu
		Gtk::RadioMenuItem * r = dynamic_cast<Gtk::RadioMenuItem *>(& m_mode.items()[new_mode]);
		assert(r);
		if (! r->get_active()) return;
		if (new_mode == mode) return;
		mode = new_mode;
		if (o_mode.get_history() != new_mode) o_mode.set_history(new_mode);
	}

	if (old_mode == MODE_SCORE) { // save scoring when mode changes
		int bt = 0;
		int wt = 0;
		int bc = 0;
		int wc = 0;
		ccgo::Marking * mk = new ccgo::Marking;
		for (int i = loc_range() - 1; i >= 0; i --) switch (terri[i]) {
		case TERRI_BLACK:
			mk->add_mark(i, ccgo::MARK_B_TERRITORY);
			bt ++;
			if (Board::get_state(i) == STATE_WHITE) bc ++;
			break;
		case TERRI_WHITE:
			mk->add_mark(i, ccgo::MARK_W_TERRITORY);
			wt ++;
			if (Board::get_state(i) == STATE_BLACK) wc ++;
			break;
		case TERRI_NEUTRAL:
			b->add_mark(i, ccgo::MARK_NONE);
			break;
		}
		where()->set_prop(ccgo::Marking::the_id(), mk);

		double komi = where()->game()->get_komi();
		double ts = bt + bc + get_b_cap() - wt - wc - get_w_cap() - komi;
		std::ostringstream s;
		s << "Score: ";
		if (ts > 0) s << "B+" << ts;
		else if (ts < 0) s << "W+" << - ts;
		else s << "tie!";
		s << " (B:" << bt << '+' << bc + get_b_cap() << " W:" << wt << '+' << wc + get_w_cap();
		if (komi >= 0) s << '+';
		s << komi << ")\n";
		tv.get_buffer()->insert(tv.get_buffer()->begin(), s.str());
	}

	assert(new_mode != MODE_INIT || game_can_init());
	assert(where());
	o_turn.set_sensitive(mode == MODE_SETUP);
	switch (mode) {
	case MODE_VIEW:
		b->set_track(gtk::Board::TRACK_NONE);
		persist_mode = MODE_VIEW;
		break;
	case MODE_INIT:
		b->set_track(gtk::Board::TRACK_BLACK);
		game_set_handi(0);
		draw_map_handle();
		break;
	case MODE_PLAY:
		b->set_track(gtk::Board::TrackMode(get_turn()));
		persist_mode = MODE_PLAY;
		break;
	case MODE_MARK:
		b->set_track(gtk::Board::TRACK_ALL);
		persist_mode = MODE_MARK;
		break;
	case MODE_SETUP:
		if (where()->down_num() || where()->get_move() != ccgo::MOVE_NONE || where() == (ccgo::Node *) where()->game()) {
			// create a setup node
			create_node();
		}
		b->set_track(gtk::Board::TRACK_ALL);
		break;
	case MODE_SCORE:
		{
			// create terri from Marking
			if (terri) delete terri;
			terri = new Terri[loc_range()];
			for (int i = loc_range() - 1; i >= 0; i --) terri[i] = ccgo::Board::TERRI_NEUTRAL;
			ccgo::Marking * mk = dynamic_cast<ccgo::Marking *>(where()->get_prop(ccgo::Marking::the_id()));
			if (mk) {
				const std::map<int, ccgo::Mark> m = mk->get_marks();
				for (std::map<int, ccgo::Mark>::const_iterator i = m.begin(); i != m.end(); i ++) {
					if (i->second == ccgo::MARK_W_TERRITORY) {
						terri[i->first] = TERRI_WHITE;
					} else if (i->second == ccgo::MARK_B_TERRITORY) {
						terri[i->first] = TERRI_BLACK;
					}
				}
			}
			do_score();
		}
		b->set_track(gtk::Board::TRACK_STONE);
		break;
	}
}

void View::game_turn_change()
{
	if (ccgo::Position::Turn(o_turn.get_history() + 1) == get_turn()) return;
	assert(mode == MODE_SETUP);
	ccgo::Position::Turn t = ccgo::Position::Turn(o_turn.get_history() + 1);

	game_set_turn(t);
	draw_map_handle();
}

void View::mark_set()
{
	if (mk_loc < 0) return;
	assert(where());

	ccgo::Marking * mk = dynamic_cast<ccgo::Marking *>(where()->get_prop(ccgo::Marking::the_id()));
	if (! mk) {
		mk = new ccgo::Marking;
		where()->set_prop(ccgo::Marking::the_id(), mk);
	}

	mk->add_mark(mk_loc, mkw.get_mark());
	b->add_mark(mk_loc, mkw.get_mark());

	ccgo::Labels * lb = dynamic_cast<ccgo::Labels *>(where()->get_prop(ccgo::Labels::the_id()));
	if (! lb) {
		lb = new ccgo::Labels;
		where()->set_prop(ccgo::Labels::the_id(), lb);
	}

	lb->add_label(mk_loc, mkw.get_label());
	b->add_label(mk_loc, mkw.get_label());
}

void View::clear_marks()
{
	if (mode == MODE_SCORE) m_mode.items()[persist_mode].activate();
	where()->set_prop(ccgo::Marking::the_id(), 0);
	for (int i = loc_range() - 1; i >= 0; i --) b->add_mark(i, ccgo::MARK_NONE);
}

void View::do_score()
{
	assert(terri);
	score_terri(terri);
	int bt = 0;
	int wt = 0;
	int bc = 0;
	int wc = 0;
	for (int i = loc_range() - 1; i >= 0; i --) switch (terri[i]) {
	case TERRI_BLACK:
		b->add_mark(i, ccgo::MARK_B_TERRITORY);
		bt ++;
		if (Board::get_state(i) == STATE_WHITE) bc ++;
		break;
	case TERRI_WHITE:
		b->add_mark(i, ccgo::MARK_W_TERRITORY);
		wt ++;
		if (Board::get_state(i) == STATE_BLACK) wc ++;
		break;
	case TERRI_NEUTRAL:
		b->add_mark(i, ccgo::MARK_NONE);
		break;
	}
	msg(DBG_INFO) << "B: territory " << bt << ", capture " << bc << "; W: territory " << wt << ", capture " << wc << '\n';

	// update score
	std::ostringstream s;
	double bs = bt + bc + get_b_cap();
	double ws = wt + wc + get_w_cap();
	double km = where()->game()->get_komi();
	if (km > 0) ws += km;
	else bs -= km;
	s << bs;
	b_cap.set_text(s.str());
	s.str("");
	s << ws;
	w_cap.set_text(s.str());
}
