/* ccgo: go/gtk/gtk.hh
 * 
 * Copyright 2004 Chun-Chung Chen <cjj@u.washington.edu>
 * 
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef GO_GTK_GTK_HH
#define GO_GTK_GTK_HH

#include "board.hh"
#include "gmap.hh"
#include <go/go.hh>

#include <gtkmm/paned.h>
#include <gtkmm/entry.h>
#include <gtkmm/button.h>
#include <gtkmm/textview.h>
#include <gtkmm/adjustment.h>
#include <gtkmm/optionmenu.h>
#include <gtkmm/radiobutton.h>
#include <gtkmm/togglebutton.h>
#include <gtkmm/fileselection.h>
#include <gtkmm/checkmenuitem.h>

namespace gtk {
	class MarkWin :
		public Gtk::Window
	{
	public:
		MarkWin();
		void show_win(int x, int y, std::string loc = "");
		void set_mark(ccgo::Mark marking);
		void set_label(std::string label);
		ccgo::Mark get_mark();
		const std::string & get_label();
	protected:
		void select_mark(ccgo::Mark marking);
		void enter_label();
	private:
		ccgo::Mark m;
		std::string l;
		std::vector<Gtk::RadioButton *> rbs;
		Gtk::Entry et;
	};

	class View :
		public Gtk::Window,
		public ccgo::Walk
	{
	public:
		View();
		~View();
		virtual void set_node(ccgo::Node * node);
	protected:
		virtual void on_realize();
		virtual bool on_configure_event(GdkEventConfigure * event);
		virtual void on_size_allocate(GtkAllocation * allocation);

		virtual void clear();
		virtual void set_state(int loc, State state);
		virtual void set_ko(const std::set<int> & ko);
	private:

		enum Mode {
			MODE_VIEW = 0,
			MODE_INIT = 1,
			MODE_PLAY = 2,
			MODE_MARK = 3,
			MODE_SETUP = 4,
			MODE_SCORE = 5,
			MODE_UNKNOWN
		};

		void toggle_mark_last();
		void toggle_show_ko();
		void set_max_depth();
		void depth_change();
		void set_alternative();
		void alternative_left();
		void alternative_right();
		void board_clicked(int loc);
		void delete_moves();
		void create_node();
		void do_old_node();
		void do_new_node();
		void paned_moved(GtkAllocation * allocation);
		void map_toggle();
		void map_hide();
		bool map_handle_draw(GdkEventExpose * event);
		void draw_map_handle();
		void select_node(ccgo::Node * node);

		void file_open();
		void file_save();

		void game_pass();
		void game_handicap(int num_stones);
		void game_mode(Mode new_mode);
		void game_turn_change();

		void mark_set();
		void clear_marks();
		void do_score();

		gtk::Board * b;
		Glib::RefPtr<Gtk::MenuItem> m_init;
		Glib::RefPtr<Gtk::CheckMenuItem> ml; // check item for marking last move
		Glib::RefPtr<Gtk::CheckMenuItem> sk; // check item for showing ko
		Gtk::Label dp; // branch depth
		Glib::RefPtr<Gtk::Adjustment> prog; // progression
		std::vector<int> path;
		Gtk::Button btL;
		Gtk::Button btR;
		Gtk::ToggleButton btM;
		Gtk::DrawingArea daM;
		Gtk::OptionMenu o_mode;
		Gtk::OptionMenu o_turn;

		MapWin mw;

		Gtk::TextView tv;
		Gtk::VPaned vp;

		int sv_w; // saved width
		int sv_h; // saved height
		int sv_p; // saved position

		Mode mode;
		Mode persist_mode;
		Gtk::Menu m_mode;

		MarkWin mkw;
		int mk_loc;

		// for scoring
		ccgo::Board::Terri * terri;

		// file selection
		Gtk::FileSelection file_win;
		ccgo::Game * my_game;

		// some more display
		Gtk::Label b_cap;
		Gtk::Label w_cap;
	};
}
#endif // GO_GTK_GTK_HH
